package edu.washington.cs.oneswarm.f2f.permissions;

import java.beans.XMLDecoder;
import java.beans.XMLEncoder;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.bouncycastle.util.encoders.Base64;
import org.gudy.azureus2.core3.disk.DiskManagerFileInfo;
import org.gudy.azureus2.core3.download.DownloadManager;
import org.gudy.azureus2.core3.download.DownloadManagerListener;
import org.gudy.azureus2.core3.global.GlobalManagerListener;
import org.gudy.azureus2.core3.util.ByteFormatter;
import org.gudy.azureus2.core3.util.HashWrapper;
import org.gudy.azureus2.core3.util.SystemProperties;
import org.gudy.azureus2.plugins.ipc.IPCException;
import org.gudy.azureus2.plugins.ipc.IPCInterface;

import com.aelitis.azureus.core.impl.AzureusCoreImpl;

import edu.washington.cs.oneswarm.f2f.Friend;
import edu.washington.cs.oneswarm.f2f.share.DownloadManagerStarter;
import edu.washington.cs.oneswarm.f2f.share.DownloadManagerStarter.DownloadManagerStartListener;

public class PermissionsDAO
{
	private static Logger				 logger								= Logger.getLogger(PermissionsDAO.class.getName());

	private static final String	 PERMISSIONS_FILE_NAME = "permissions.xml";

	private static final String	 PERMISSIONS_PATH			= SystemProperties.getUserPath()
																													+ File.separator
																													+ PERMISSIONS_FILE_NAME;

	private static PermissionsDAO inst									= new PermissionsDAO();

	/**
	 * Hard state. This needs to be externally synchronized because the XML encoder can't deal with synchronized map types. 
	 */
	Map<String, List<GroupBean>>	infohash_to_groups		= new HashMap<String, List<GroupBean>>();

	Map<String, GroupBean>				groupname_to_group		= new HashMap<String, GroupBean>();

	/**/

	private IPCInterface					f2fIpc								= null;

	public static final PermissionsDAO get() {
		return inst;
	}

	protected PermissionsDAO() {
	}

	public synchronized void init(IPCInterface f2f) {
		if (f2fIpc == null) {
			f2fIpc = f2f;
			load_permissions();
		}
	}

	/**
	 * This is a hack, but is needed to deal with load order issues. 
	 */
	public synchronized void f2fInitialized() {
		AzureusCoreImpl.getSingleton().getGlobalManager().addListener(
				new GlobalManagerListener() {
					public void destroyInitiated() {
					}

					public void destroyed() {
					}

					public void downloadManagerAdded(DownloadManager dm) {

						try {
							String hexHash = ByteFormatter.encodeString(dm.getTorrent().getHash());
							List<GroupBean> groups = PermissionsDAO.this.getGroupsForHash(hexHash);
							boolean any_friend = false;
							for (GroupBean gb : groups) {
								if (gb.equals(GroupBean.PUBLIC) == false) {
									any_friend = true;
								}
							}

							logger.fine("setTorrentPrivacy: " + hexHash + " pub?: "
									+ Boolean.toString(groups.contains(GroupBean.PUBLIC))
									+ " f2f?: " + Boolean.toString(any_friend));

							try {
								f2fIpc.invoke("setTorrentPrivacy", new Object[] {
									ByteFormatter.decodeString(hexHash),
									groups.contains(GroupBean.PUBLIC),
									any_friend
								});
							} catch (Exception e) {
								e.printStackTrace();
							}

							logger.finer("done setTorrentPrivacy");
						} catch (Exception e) {
							e.printStackTrace();
						}

					}

					public void downloadManagerRemoved(DownloadManager dm) {
						try {
							if (infohash_to_groups.remove(ByteFormatter.encodeString(dm.getTorrent().getHash())) != null) {
								logger.fine("removed from perms: " + dm.getDisplayName());
							} else {
								logger.warning("couldn't find perms for removed DL manager: "
										+ dm.getDisplayName());
							}
							save_permissions();
						} catch (Exception e) {
							e.printStackTrace();
						}
					}

					public void seedingStatusChanged(boolean seeding_only_mode) {
					}
				});
	}

	private void load_permissions() {
		try {
			logger.info("loading permissions...");

			XMLDecoder decoder = new XMLDecoder(new BufferedInputStream(
					new FileInputStream(PERMISSIONS_PATH)));

			infohash_to_groups = (Map<String, List<GroupBean>>) decoder.readObject();
			groupname_to_group = (Map<String, GroupBean>) decoder.readObject();

			/**
			 * Sanity check stored values, eliminate null groups, etc. 
			 */
			for (String groupName : groupname_to_group.keySet()) {
				if (groupname_to_group.get(groupName) == null) {
					groupname_to_group.remove(groupName);
					logger.fine("removed groupname with null group: " + groupName);
				}
			}
			for (String infohash : infohash_to_groups.keySet()) {
				if (infohash_to_groups.get(infohash) == null) {
					infohash_to_groups.remove(infohash);
					logger.fine("removed infohash to group mapping with null groups: "
							+ infohash);
				}
			}

			if (logger.isLoggable(Level.FINEST)) {
				for (String hash : infohash_to_groups.keySet()) {
					logger.finest(hash + " has: ");
					for (GroupBean g : infohash_to_groups.get(hash)) {
						logger.finest("\t" + g.getGroupName());
					}
				}
			}

			//			/**
			//			 * TODO: HACK to fix previous errors, remove after a version or two 
			//			 */
			//			for( DownloadManager dm : (List<DownloadManager>)AzureusCoreImpl.getSingleton().getGlobalManager().getDownloadManagers() ) {
			//				if( getGroupsForHash(ByteFormatter.encodeString(dm.getTorrent().getHash())) == null ) {
			//					logger.warning("no permissions for existing swarm: " + dm.getDisplayName() + " setting friends only default...");
			//					List<GroupBean> def = new ArrayList<GroupBean>();
			//					def.add(GroupBean.ALL_FRIENDS);
			//					setGroupsForHash(ByteFormatter.encodeString(dm.getTorrent().getHash()), def);
			//				}
			//			}

			refresh_friend_groups();

			logger.info("loaded existing permissions for "
					+ infohash_to_groups.size());

		} catch (Exception e) {
			logger.warning("couldn't load permissions: " + e.toString());

			firstload();
		}
	}

	public synchronized void refresh_friend_groups() {
		logger.fine("refresh friend groups");
		try {
			List<Friend> raw_friends = (List<Friend>) f2fIpc.invoke("getFriends",
					new Object[0]);
			List<Friend> friends = new ArrayList<Friend>();

			for (Friend f : raw_friends) {
				if (f.isBlocked() == false && f.isCanSeeFileList()) {
					friends.add(f);
				}
			}

			logger.fine("iterating over: " + friends.size() + " friends");

			Set<String> not_covered = new HashSet<String>();
			for (String s : groupname_to_group.keySet()) {

				if (groupname_to_group.get(s) == null) {
					logger.warning("groupname points to null group: " + s);
					continue;
				}

				/**
				 * We're only considering user groups here.
				 */
				if (groupname_to_group.get(s).isUserGroup()) {
					not_covered.add(s);
				}
			}

			// First, add any new groups that have been added
			for (Friend f : friends) {
				logger.finer("considering: " + f.getNick());
				if (groupname_to_group.containsKey(f.getNick()) == false) {
					List<String> keyList = new ArrayList<String>();
					keyList.add(new String(Base64.encode(f.getPublicKey())));
					logger.finer("adding new group: " + f.getNick());
					addGroup(f.getNick(), keyList, true);
				} else {
					not_covered.remove(f.getNick());
				}
			}
			// Next, remove any old groups that are not covered by the set of
			// current groups
			for (String deadGroupName : not_covered) {
				logger.finer("removing dead group: " + deadGroupName);
				try {
					removeGroup(deadGroupName, true); // need special method
					// since this is a user
					// group
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
		} catch (IPCException e) {
			logger.warning("IPC error: " + e.toString() + " (check stderr)");
			e.printStackTrace();
		}
	}

	public synchronized void renameGroup(String old, String neu) {
		try {
			if (old.equals(neu)) {
				logger.warning("Skipping rename, old = neu: " + old);
				return;
			}

			GroupBean g = groupname_to_group.remove(old);
			if (g == null) {
				logger.severe("trying to rename nonexistant group! old: " + old
						+ " neu: " + neu);
				//(new Exception()).printStackTrace();
				return;
			}
			groupname_to_group.put(neu, g);
			g.setGroupName(neu);

			/**
			 * This may or may not be necessary, but to be on the safe side...
			 */
			for (List<GroupBean> l : infohash_to_groups.values()) {
				for (GroupBean group : l) {
					if (group.isUserGroup() && group.getGroupName().equals(old)) {
						group.setGroupName(neu);
						logger.fine("updating values group name on rename. it is necessary!");
					}
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	private void firstload() {
		groupname_to_group.put(GroupBean.ALL_FRIENDS.getGroupName(),
				GroupBean.ALL_FRIENDS);
		groupname_to_group.put(GroupBean.PUBLIC.getGroupName(), GroupBean.PUBLIC);

		try {
			logger.fine("calling getfriends");
			List<Friend> friends = (List<Friend>) f2fIpc.invoke("getFriends",
					new Object[0]);
			for (Friend f : friends) {
				List<String> keyList = new ArrayList<String>();
				keyList.add(new String(Base64.encode(f.getPublicKey())));
				addGroup(f.getNick(), keyList, true);
			}
			logger.fine("added: " + friends.size());
		} catch (Exception e) {
			logger.warning("couldn't get friends list (plugin not loaded yet?)");
			e.printStackTrace();
		}
	}

	public void refreshFileLists() {
		try {
			logger.fine("refresh file lists in f2f...");
			f2fIpc.invoke("refreshFileLists", new Object[0]);
			logger.fine("done refreshFileLists()");
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	private synchronized void save_permissions() {
		logger.fine("save permissions");
		try {
			XMLEncoder encoder = new XMLEncoder(new BufferedOutputStream(
					new FileOutputStream(PERMISSIONS_PATH)));
			encoder.writeObject(infohash_to_groups);
			encoder.writeObject(groupname_to_group);
			encoder.close();
			logger.fine("done");
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	public synchronized void addGroup(String inName, List<String> userKeys,
			boolean isUserGroup) {
		if (groupname_to_group.containsKey(inName)) {
			logger.warning("Group already exists: " + inName);
			return;
		}

		GroupBean g = new GroupBean();
		g.setGroupName(inName);
		g.setMemberKeys(userKeys);
		g.setUserGroup(isUserGroup);
		groupname_to_group.put(inName, g);
		save_permissions();
		logger.fine("added new group: " + inName);
	}

	public synchronized void removeGroup(String inName) throws IOException {
		removeGroup(inName, false);
	}

	public synchronized void removeGroup(String inName, boolean removeUser)
			throws IOException {
		if (groupname_to_group.containsKey(inName) == false) {
			throw new IOException("tried to remove nonexistent group");
		}

		if (inName.equals(GroupBean.ALL_FRIENDS.getGroupName())
				|| inName.equals(GroupBean.PUBLIC.getGroupName())) {
			throw new IOException("cannot remove reserved group: " + inName);
		}

		if (removeUser == false) {
			if (groupname_to_group.get(inName).isUserGroup()) {
				throw new IOException("cannot remove reserved group: " + inName);
			}
		}

		groupname_to_group.remove(inName);
		for (List<GroupBean> groups : infohash_to_groups.values()) {
			List<Integer> toRemove = new ArrayList<Integer>();
			for (int i = 0; i < groups.size(); i++) {
				if (groups.get(i).getGroupName().equals(inName))
					toRemove.add(i);
			}
			for (int i = toRemove.size() - 1; i >= 0; i--) {
				groups.remove(i);
			}
		}
		save_permissions();

		logger.fine("successfully removed: " + inName);
	}

	public synchronized boolean hasPublicPermission(byte[] infoHash) {
		String inHexHash = ByteFormatter.encodeString(infoHash);
		List<GroupBean> allowedGroups = getGroupsForHash(inHexHash);
		if (allowedGroups == null) {
			return false;
		}
		return allowedGroups.contains(GroupBean.PUBLIC);
	}

	public synchronized boolean hasAllFriendsPermission(byte[] infoHash) {
		String inHexHash = ByteFormatter.encodeString(infoHash);
		List<GroupBean> allowedGroups = getGroupsForHash(inHexHash);
		if (allowedGroups == null) {
			return false;
		}
		return allowedGroups.contains(GroupBean.ALL_FRIENDS);
	}

	public synchronized List<GroupBean> getGroupsForHash(String inHexHash) {
		if (infohash_to_groups.containsKey(inHexHash) == false) {
			logger.fine("couldn't find perms for hash: " + inHexHash
					+ " - default is no permissions");
			//			// defaults to whatever the all friends / public settings are in the
			//			// core for the swarm
			//			DownloadManager dm = AzureusCoreImpl.getSingleton().getGlobalManager().getDownloadManager(new HashWrapper(ByteFormatter.decodeString(inHexHash)));
			//			if (dm == null) {
			//				logger.warning("trying to get groups for hash with no download manager!");
			//				(new Exception()).printStackTrace();
			//				List<GroupBean> nothing = new ArrayList<GroupBean>(0);
			//				return nothing;
			//			} else {
			//				List<GroupBean> defaults = new ArrayList<GroupBean>();
			//				for (String s : dm.getDownloadState().getNetworks()) {
			//					if (s.equals(AENetworkClassifier.AT_OSF2F)) {
			//						defaults.add(GroupBean.ALL_FRIENDS);
			//						logger.finer("adding f2f to defaults");
			//					} else if (s.equals(AENetworkClassifier.AT_PUBLIC)) {
			//						defaults.add(GroupBean.PUBLIC);
			//						logger.finer("adding public to defaults");
			//					} else {
			//						logger.finer("unknown network: " + s);
			//					}
			//				}
			//				infohash_to_groups.put(inHexHash, defaults);
			//			}

			return new ArrayList<GroupBean>();
		}

		return infohash_to_groups.get(inHexHash);
	}

	public synchronized void setGroupsForHash(String inHexHash,
			final List<GroupBean> inGroups) {

		logger.fine("setGroupsForHash: " + inHexHash + " (" + inGroups.size()
				+ " groups)");

		List<GroupBean> existing = getGroupsForHash(inHexHash);
		boolean _started_with_public = false;
		if (existing != null) {
			if (existing.contains(GroupBean.PUBLIC)) {
				logger.fine(inHexHash + " started with public");
				_started_with_public = true;
			}
		}
		final boolean started_with_public = _started_with_public;

		logger.finest(inHexHash + " has " + existing.size()
				+ " existing groups. New groups: ");
		for (GroupBean g : inGroups) {
			logger.finest(g.getGroupName());
		}

		infohash_to_groups.put(inHexHash, inGroups);

		/**
		 * If this includes either of the special groups, we need to update
		 * these ourselves immediately (these are not on demand)
		 */
		boolean any_friend = false;
		for (GroupBean g : inGroups) {
			if (g.equals(GroupBean.PUBLIC) == false) {
				any_friend = true;
			}
		}

		/*
		 * sync up the enabled networks
		 * 
		 */
		final DownloadManager dm = AzureusCoreImpl.getSingleton().getGlobalManager().getDownloadManager(
				new HashWrapper(ByteFormatter.decodeString(inHexHash)));
		if (dm != null) {

			logger.fine("setTorrentPrivacy: " + inHexHash + " pub?: "
					+ Boolean.toString(inGroups.contains(GroupBean.PUBLIC)) + " f2f?: "
					+ Boolean.toString(any_friend));

			try {
				f2fIpc.invoke("setTorrentPrivacy", new Object[] {
					ByteFormatter.decodeString(inHexHash),
					inGroups.contains(GroupBean.PUBLIC),
					any_friend
				});
			} catch (Exception e) {
				e.printStackTrace();
			}

			logger.finer("done setTorrentPrivacy");
		}
		/**
		 * If we add/remove the public network we need to restart the torrent
		 */
		if (started_with_public != inGroups.contains(GroupBean.PUBLIC)) {
			logger.fine("changing public visibility requires stop/start");
			if (dm != null) {
				if (dm.getState() != DownloadManager.STATE_STOPPED
						|| dm.getState() != DownloadManager.STATE_STOPPING) {
					dm.addListener(new DownloadManagerListener() {
						public void completionChanged(DownloadManager manager,
								boolean completed) {
						}

						public void downloadComplete(DownloadManager manager) {
						}

						public void filePriorityChanged(DownloadManager download,
								DiskManagerFileInfo file) {
						}

						public void positionChanged(DownloadManager download,
								int oldPosition, int newPosition) {
						}

						public void stateChanged(final DownloadManager manager, int state) {
							if (state == DownloadManager.STATE_QUEUED) {
								DownloadManagerStarter.startDownload(manager,
										new DownloadManagerStartListener() {
											public void downloadStarted() {
												if (started_with_public == false
														&& inGroups.contains(GroupBean.PUBLIC) == true) {
													logger.finer("public network added, forcing tracker update");
													manager.getTrackerClient().update(true);
												}
											}
										});
							}
						}
					});
					logger.fine("stopping download: " + dm.getDisplayName());
					dm.stopIt(DownloadManager.STATE_QUEUED, false, false);
				}
			}
		}

		refreshFileLists();
		save_permissions();
	}

	public synchronized GroupBean getGroup(String inGroupName) {
		return groupname_to_group.get(inGroupName);
	}

	public synchronized Collection<GroupBean> getAllGroups() {
		return groupname_to_group.values();
	}

	public boolean hasPermissions(String inBase64Key, byte[] inSwarmHash) {
		return hasPermissions(Base64.decode(inBase64Key), inSwarmHash);
	}

	public synchronized void updateOrCreateGroup(String inName,
			List<String> inMembers) {
		GroupBean g = groupname_to_group.get(inName);
		if (g == null) {
			g = new GroupBean();
			g.setUserGroup(false);
			g.setGroupName(inName);
			g.setMemberKeys(inMembers);
			groupname_to_group.put(inName, g);
		}

		g.memberKeys = inMembers;
		save_permissions();
		refreshFileLists();
	}

	public synchronized boolean hasPermissions(byte[] inKey, byte[] inSwarmHash) {
		String hashStr = ByteFormatter.encodeString(inSwarmHash);

		if (infohash_to_groups.get(hashStr) == null) {
			logger.finer("permissions request for unknown swarm: " + hashStr);
			return false;
		}
		String base64Key = new String(Base64.encode(inKey));
		for (GroupBean b : infohash_to_groups.get(hashStr)) {
			if (b.equals(GroupBean.ALL_FRIENDS)) {
				return true;
			}
			if (b.getMemberKeys().contains(base64Key)) {
				return true;
			}
		}
		return false;
	}

	public static final void main(String[] args) {

	}
}
