package edu.washington.cs.oneswarm.f2f.multisource;

import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;

import org.gudy.azureus2.core3.disk.DiskManagerFileInfo;
import org.gudy.azureus2.core3.download.DownloadManager;
import org.gudy.azureus2.core3.download.DownloadManagerListener;
import org.gudy.azureus2.core3.global.GlobalManagerListener;
import org.gudy.azureus2.core3.torrent.TOTorrent;
import org.gudy.azureus2.core3.torrent.TOTorrentException;
import org.gudy.azureus2.core3.util.HashWrapper;

import com.aelitis.azureus.core.AzureusCore;
import com.aelitis.azureus.core.impl.AzureusCoreImpl;

import edu.washington.cs.oneswarm.f2f.FileListFile;
import edu.washington.cs.oneswarm.f2f.multisource.Sha1PieceRequestTranslator.PieceTranslationExcetion;
import edu.washington.cs.oneswarm.f2f.share.DownloadManagerStarter;
import edu.washington.cs.oneswarm.f2f.share.DownloadManagerStarter.DownloadManagerStartListener;

public class Sha1DownloadManager {
	private final static Logger logger = Logger.getLogger(Sha1DownloadManager.class.getName());
	private final AzureusCore core;
	private Set<Sha1Peer> currentPeers = new HashSet<Sha1Peer>();

	public Sha1DownloadManager() {
		core = AzureusCoreImpl.getSingleton();
		core.getGlobalManager().addListener(new GlobalManagerListener() {
			public void seedingStatusChanged(boolean seedingOnlyMode) {
			}

			public void downloadManagerRemoved(DownloadManager dm) {
			}

			public void downloadManagerAdded(DownloadManager dm) {
				// basically we want to check if we need to add peers each time
				// a download starts
				dm.addListener(new DownloadManagerListener() {
					public void stateChanged(DownloadManager manager, int state) {
						if (manager.getState() == DownloadManager.STATE_DOWNLOADING) {
							logger.fine("download manager state changed to downloading, checking for sha1 matches: " + manager.getDisplayName());
							try {
								downloadManagerDownloading(manager);
							} catch (TOTorrentException e) {
								// TODO Auto-generated catch block
								e.printStackTrace();
							}
						}
					}

					public void positionChanged(DownloadManager download, int oldPosition, int newPosition) {
					}

					public void filePriorityChanged(DownloadManager download, DiskManagerFileInfo file) {
					}

					public void downloadComplete(DownloadManager manager) {
					}

					public void completionChanged(DownloadManager manager, boolean bCompleted) {
					}
				});
			}

			public void destroyed() {
			}

			public void destroyInitiated() {
			}
		});
	}

	@SuppressWarnings("unchecked")
	private void downloadManagerDownloading(final DownloadManager download) throws TOTorrentException {
		if (download.isDownloadComplete(false)) {
			logger.fine("skipping download manager, already completed...");
			return;
		}
		List<DownloadManager> downloadManagers = core.getGlobalManager().getDownloadManagers();

		// create a library of sha1->download manager mappings
		logger.finer("creating sha1 map");
		HashMap<HashWrapper, List<DownloadManager>> sha1Mappings = new HashMap<HashWrapper, List<DownloadManager>>();
		for (DownloadManager d : downloadManagers) {
			if (!download.equals(d)) {
				TOTorrent t = d.getTorrent();
				List<HashWrapper> sha1Found = getSha1FromTorrent(t);
				/*
				 * add everything we found
				 */
				for (HashWrapper sha1 : sha1Found) {
					if (!sha1Mappings.containsKey(sha1)) {
						sha1Mappings.put(sha1, new LinkedList<DownloadManager>());
					}
					sha1Mappings.get(sha1).add(d);
				}
			}
		}

		/*
		 * now, scan all files in the current torrent and check for matches
		 */
		logger.finest("scanning for matches");
		List<HashWrapper> torrentSha1s = getSha1FromTorrent(download.getTorrent());
		final HashSet<DownloadManager> interestingDms = new HashSet<DownloadManager>();
		for (HashWrapper t : torrentSha1s) {
			List<DownloadManager> matchingDms = sha1Mappings.get(t);
			if (matchingDms != null) {
				interestingDms.addAll(matchingDms);
			}
		}

		/*
		 * last, we have a set of interesting dms,
		 */
		logger.fine("found " + interestingDms.size() + " interesting download managers, adding sha1 peers");
		DownloadManagerStarter.startDownload(download, new DownloadManagerStartListener() {
			public void downloadStarted() {
				for (final DownloadManager srcDm : interestingDms) {
					DownloadManagerStarter.startDownload(srcDm, new DownloadManagerStartListener() {
						public void downloadStarted() {
							addPeer(srcDm, download);
							addPeer(download, srcDm);
						}
					});
				}
			}
		});

	}

	private void addPeer(DownloadManager source, final DownloadManager destination) {
		if (destination.isDownloadComplete(false)) {
			return;
		}
		boolean add = false;
		try {
			Sha1Peer peer = new Sha1Peer(this, source, destination);

			synchronized (currentPeers) {
				logger.finer("got request to add peer, currently active=" + currentPeers.size());
				if (!currentPeers.contains(peer)) {
					currentPeers.add(peer);
					add = true;
				}
			}
			if (add) {
				logger.finer("adding peer");
				destination.getPeerManager().addPeer(peer);
			} else {
				logger.finer("not adding peer (already exists)");
			}
		} catch (PieceTranslationExcetion e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	@SuppressWarnings("unchecked")
	private static List<HashWrapper> getSha1FromTorrent(TOTorrent t) throws TOTorrentException {

		Map torrentMap = (Map) t.serialiseToMap().get("info");
		List<HashWrapper> sha1Found = new LinkedList<HashWrapper>();
		if (t.isSimpleTorrent()) {
			HashWrapper torrentSha1 = getSha1FromTorrentMap(torrentMap);
			if (torrentSha1 != null) {
				sha1Found.add(torrentSha1);
			}
		} else {
			Object filePropertiesObj = torrentMap.get("files");
			if (filePropertiesObj == null || !(filePropertiesObj instanceof List)) {
				return sha1Found;
			}

			List fileProperties = (List) filePropertiesObj;
			if (fileProperties != null) {
				for (Object pObj : fileProperties) {
					if (pObj != null && pObj instanceof Map) {
						Map pMap = (Map) pObj;
						HashWrapper torrentSha1 = getSha1FromTorrentMap(pMap);
						if (torrentSha1 != null) {
							sha1Found.add(torrentSha1);
						}
					}
				}
			}
		}
		return sha1Found;
	}

	@SuppressWarnings("unchecked")
	private static HashWrapper getSha1FromTorrentMap(Map torrentMap) {
		if (torrentMap.containsKey(FileListFile.KEY_SHA1_HASH)) {
			return new HashWrapper((byte[]) torrentMap.get(FileListFile.KEY_SHA1_HASH));
		} else {
			return null;
		}
	}

	void peerClosed(Sha1Peer sha1Peer) {
		synchronized (currentPeers) {
			currentPeers.remove(sha1Peer);
		}

	}

}
